import inspect
import os
import sys

# add parent directory to be able to import 'src'
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)



from pip._vendor import pkg_resources


def get_version(package):
    package = package.lower()
    return next(
        (
            p.version
            for p in pkg_resources.working_set  # noqa: F821
            if p.project_name.lower() == package
        ),
        "No match",
    )


__version__ = get_version('whathappened')

del get_version
del pkg_resources



import pytest

from whathappened import changelog as cl


def dummy_version(ref, date, breaking, feature, fix, num_commits):
    version = cl.Version(ref, date)
    version.breaking = breaking
    version.feature = feature
    version.fix = fix
    version.commits = [None for each in range(num_commits)]

    return version


def assert_version(version_a, version_b):
    assert version_a.ref == version_b.ref
    assert version_a.date == version_b.date
    assert version_a.breaking == version_b.breaking
    assert version_a.feature == version_b.feature
    assert version_a.fix == version_b.fix
    assert len(version_a.commits) == len(version_b.commits)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (
            "Break feat(read me): specify expected message format",
            {
                'description': 'specify expected message format',
                'type': 'feat',
                'scope': 'read me',
                'is_breaking': True,
                'is_feature': True,
                'is_fix': False,
            },
        ),
        (
            "docs(README.md): specify expected message format",
            {
                'description': 'specify expected message format',
                'type': 'docs',
                'scope': 'README.md',
                'is_breaking': False,
                'is_feature': False,
                'is_fix': False,
            },
        ),
        (
            "BREAKING fix (code): repair things",
            {
                'description': 'repair things',
                'type': 'fix',
                'scope': 'code',
                'is_breaking': True,
                'is_feature': False,
                'is_fix': True,
            },
        ),
        (
            "breaking fix: repair things",
            {
                'description': 'repair things',
                'type': 'fix',
                'scope': None,
                'is_breaking': True,
                'is_feature': False,
                'is_fix': True,
            },
        ),
        (
            "fix: add inspiration",
            {
                'description': 'add inspiration',
                'type': 'fix',
                'scope': None,
                'is_breaking': False,
                'is_feature': False,
                'is_fix': True,
            },
        ),
        (
            "docs (readme): add badges",
            {
                'description': 'add badges',
                'type': 'docs',
                'scope': 'readme',
                'is_breaking': False,
                'is_feature': False,
                'is_fix': False,
            },
        ),
        (
            "build(actions): create python-app.yml for github action's",
            {
                'description': 'create python-app.yml for github action\'s',
                'type': 'build',
                'scope': 'actions',
                'is_breaking': False,
                'is_feature': False,
                'is_fix': False,
            },
        ),
        (
            "README.md: correct typo",
            {
                'description': 'README.md: correct typo',
                'type': 'other',
                'scope': None,
                'is_breaking': False,
                'is_feature': False,
                'is_fix': False,
            },
        ),
        (
            "Initial commit",
            {
                'description': 'Initial commit',
                'type': 'other',
                'scope': None,
                'is_breaking': False,
                'is_feature': False,
                'is_fix': False,
            },
        ),
    ],
)
def test_commit_title_parsing(test_input, expected):
    commit = cl.Commit({'hash': '00000', 'title': test_input})

    commit_attr = {}
    for attr in expected.keys():
        commit_attr[attr] = getattr(commit, attr)

    print(commit_attr)

    assert commit_attr == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (
            [
                {
                    'hash': 'e324c324df48a76113ad9b3c0887f161324046e4',
                    'tags': ['v0.1.1'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 17:30:25 2020 +0200',
                    'message': '',
                    'title': 'breaking feat(readme): specify expected message format',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'breaking fix (code): repair things',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'breaking fix: repair things',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'fix: add inspiration',
                },
                {
                    'hash': 'f60445bba0ac48e12ce6be5526644037234ae500',
                    'tags': ['v0.0.1'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:31 2020 +0200',
                    'message': '',
                    'title': 'docs (readme): add badges',
                },
                {
                    'hash': '9e57ba91f54244af913931c017480a39605c15f9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 13:55:04 2020 +0200',
                    'message': (
                        'Setup Python 3.6 on ubuntu-latest\n'
                        'Install pipenv and dependencies\n'
                        'Test'
                    ),
                    'title': 'build(actions): create python-app.yml for github actions',
                },
                {
                    'hash': '4094d22846daea951c4fe0d74abb2a798e9a3404',
                    'tags': ['v0.0.0'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 13:19:28 2020 +0200',
                    'message': '',
                    'title': 'Initial commit',
                },
            ],
            [
                dummy_version(
                    'v0.1.1',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=3,
                    feature=1,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    'v0.0.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
                dummy_version(
                    'v0.0.0',
                    'Sat Oct 17 13:19:28 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=1,
                ),
            ],
        )
    ],
)
def test_compile_log_versions(test_input, expected):
    commits = test_input
    versions = cl.compile_log(commits)

    for version, expectation in zip(versions, expected):
        assert_version(version, expectation)


@pytest.mark.parametrize(
    "versions, expected",
    [
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=3,
                    feature=1,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    '0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
                dummy_version(
                    '0.0.0',
                    'Sat Oct 17 13:19:28 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=1,
                ),
            ],
            "0.2.0",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=1,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    '0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "0.1.2",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    '0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "0.1.2",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=4,
                ),
                dummy_version(
                    '0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "0.1.1",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=1,
                    feature=1,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "2.0.0",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=1,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "1.1.0",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "1.0.1",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "1.0.0",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=1,
                    feature=1,
                    fix=1,
                    num_commits=4,
                ),
            ],
            "0.1.0",
        ),
    ],
)
def test_calculate_next(versions, expected):
    next_version = cl.calculate_next(versions)

    assert next_version == expected


@pytest.mark.parametrize(
    "versions, prefix, expected",
    [
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "",
            "1.0.1",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    'version1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "version",
            "version1.0.1",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    'v1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "v",
            "v1.0.1",
        ),
    ],
)
def test_calculate_next_prefixes_success(versions, prefix, expected):
    next_version = cl.calculate_next(versions, prefix=prefix)

    assert next_version == expected


@pytest.mark.parametrize(
    "versions, prefix",
    [
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    '1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "v",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    'version1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "version_",
        ),
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=1,
                    num_commits=4,
                ),
                dummy_version(
                    'v1.0.0',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
            ],
            "version",
        ),
    ],
)
def test_calculate_next_prefixes_failure(versions, prefix):
    with pytest.raises(cl.VersionFormatException):
        cl.calculate_next(versions, prefix=prefix)


@pytest.mark.parametrize(
    "versions",
    [
        [
            dummy_version(
                'HEAD',
                'Sat Oct 17 17:30:25 2020 +0200',
                breaking=0,
                feature=0,
                fix=1,
                num_commits=4,
            ),
            dummy_version(
                '01.0.0',
                'Sat Oct 17 15:00:31 2020 +0200',
                breaking=0,
                feature=0,
                fix=0,
                num_commits=2,
            ),
        ],
        [
            dummy_version(
                'HEAD',
                'Sat Oct 17 17:30:25 2020 +0200',
                breaking=0,
                feature=0,
                fix=1,
                num_commits=4,
            ),
            dummy_version(
                '1.0.',
                'Sat Oct 17 15:00:31 2020 +0200',
                breaking=0,
                feature=0,
                fix=0,
                num_commits=2,
            ),
        ],
        [
            dummy_version(
                'HEAD',
                'Sat Oct 17 17:30:25 2020 +0200',
                breaking=0,
                feature=0,
                fix=1,
                num_commits=4,
            ),
            dummy_version(
                '1.0',
                'Sat Oct 17 15:00:31 2020 +0200',
                breaking=0,
                feature=0,
                fix=0,
                num_commits=2,
            ),
        ],
        [
            dummy_version(
                'HEAD',
                'Sat Oct 17 17:30:25 2020 +0200',
                breaking=0,
                feature=0,
                fix=1,
                num_commits=4,
            ),
            dummy_version(
                '1',
                'Sat Oct 17 15:00:31 2020 +0200',
                breaking=0,
                feature=0,
                fix=0,
                num_commits=2,
            ),
        ],
    ],
)
def test_calculate_next_version_string_failure(versions):
    with pytest.raises(cl.VersionFormatException):
        result = cl.calculate_next(versions)

        print(result)


@pytest.mark.parametrize(
    "versions, prefix, expected",
    [
        (
            [
                dummy_version(
                    'HEAD',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=3,
                    feature=1,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    'v0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
                dummy_version(
                    'v0.0.0',
                    'Sat Oct 17 13:19:28 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=1,
                ),
            ],
            "v",
            [
                dummy_version(
                    'v0.2.0',
                    'Sat Oct 17 17:30:25 2020 +0200',
                    breaking=3,
                    feature=1,
                    fix=3,
                    num_commits=4,
                ),
                dummy_version(
                    'v0.1.1',
                    'Sat Oct 17 15:00:31 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=2,
                ),
                dummy_version(
                    'v0.0.0',
                    'Sat Oct 17 13:19:28 2020 +0200',
                    breaking=0,
                    feature=0,
                    fix=0,
                    num_commits=1,
                ),
            ],
        )
    ],
)
def test_update_latest_version(versions, prefix, expected):
    versions = cl.update_latest_version(versions, prefix=prefix)

    for version, expectation in zip(versions, expected):
        assert_version(version, expectation)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (
            [
                {
                    'hash': 'e324c324df48a76113ad9b3c0887f161324046e4',
                    'tags': ['v0.1.1'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 17:30:25 2020 +0200',
                    'message': '',
                    'title': 'breaking feat(readme): specify expected message format',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'breaking fix (code): repair things',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'breaking fix: repair things',
                },
                {
                    'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:48 2020 +0200',
                    'message': '',
                    'title': 'fix: add inspiration',
                },
                {
                    'hash': 'f60445bba0ac48e12ce6be5526644037234ae500',
                    'tags': ['v0.0.1'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 15:00:31 2020 +0200',
                    'message': '',
                    'title': 'docs (readme): add badges',
                },
                {
                    'hash': '9e57ba91f54244af913931c017480a39605c15f9',
                    'tags': [],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 13:55:04 2020 +0200',
                    'message': (
                        'Setup Python 3.6 on ubuntu-latest\n'
                        'Install pipenv and dependencies\n'
                        'Test'
                    ),
                    'title': 'build(actions): create python-app.yml for github actions',
                },
                {
                    'hash': '4094d22846daea951c4fe0d74abb2a798e9a3404',
                    'tags': ['v0.0.0'],
                    'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
                    'date': 'Sat Oct 17 13:19:28 2020 +0200',
                    'message': '',
                    'title': 'Initial commit',
                },
            ],
            """# Changelog

## v0.1.1 (2020-10-17)

### Features

* Readme - specify expected message format [BREAKING]

### Fixes

* Add inspiration
* Repair things [BREAKING]
* Code - repair things [BREAKING]


## v0.0.1 (2020-10-17)

### Docs

* Readme - add badges


## v0.0.0 (2020-10-17)

### Other

* Initial commit
""",
        )
    ],
)
def test_changelog(test_input, expected):
    commits = test_input
    versions = cl.compile_log(commits)
    log = cl.format_log(versions)

    print(log)

    assert log == expected



import re

from datetime import datetime
from itertools import groupby

from git_commits import get_commits

# see http://semver.org
semver_regex = re.compile(
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
    r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


class Version:
    def __init__(self, ref, date):
        self.ref = ref
        self.date = (
            datetime.strptime(date, "%a %b %d %H:%M:%S %Y %z")
            if isinstance(date, str)
            else date
        )  # str input format Wed Apr 22 18:58:54 2020 +0200

        self.breaking = 0
        self.feature = 0
        self.fix = 0

        self.commits = []

    def __repr__(self):
        return (
            f"Version({self.ref},"
            f" {self.date},"
            f" breaking={self.breaking},"
            f" feature={self.feature},"
            f" fix={self.fix},"
            f" num_commits={len(self.commits)})"
        )


class Commit:
    def __init__(self, commit_dict):
        self.commit_dict = commit_dict

    def __getattr__(self, name):

        try:
            return self.commit_dict[name]
        except KeyError:
            if name == 'header':
                return self.title.split(': ', 1)[0]
            elif name == 'description':
                return (
                    self.title.split(': ', 1)[1] if ': ' in self.title else self.title
                )
            elif name == 'is_breaking':
                return self.header.lower().startswith('break')
            elif name == 'type':
                return self.header.lower().split(' ', 1)[1 if self.is_breaking else 0]
            elif name == 'scope':
                return self.header.split('(')[1][:-1] if '(' in self.header else None
            elif name == 'is_feature':
                return "feat" in self.header.lower()
            elif name == 'is_fix':
                return "fix" in self.header.lower()
            else:
                raise AttributeError(f"Attribute '{name}' not found in class Commit")

    def __repr__(self):
        return f"{self.hash[:8]} {self.title})"


def sentence(string):
    try:
        return string[0].upper() + string[1:]
    except IndexError:
        # zero-length string
        return string


def retrieve_current_version(versions, prefix=''):
    global semver_regex

    last_version = None

    for version in versions:
        if version.ref != 'HEAD':
            last_version = version.ref
            break

    if last_version is None:
        return f"{prefix}0.1.0"
    else:
        result = semver_regex.match(last_version[len(prefix) :])
        major, minor, patch = (
            int(result.group('major')),
            int(result.group('minor')),
            int(result.group('patch')),
        )

        # print(major, minor, patch)

        this_version = versions[0]

        if major == 0:
            # this is a development release and only receives minor and patch increments
            if this_version.breaking > 0:
                return f"{prefix}{major}.{minor+1}.0"
            elif this_version.feature > 0 or this_version.fix > 0:
                return f"{prefix}{major}.{minor}.{patch+1}"
            else:
                # no api changes
                return f"{prefix}{major}.{minor}.{patch}"

        else:
            # this is production release and receives major, minor, and patch increments
            if this_version.breaking > 0:
                return f"{prefix}{major+1}.0.0"
            elif this_version.feature > 0:
                return f"{prefix}{major}.{minor+1}.0"
            elif this_version.fix > 0:
                return f"{prefix}{major}.{minor}.{patch+1}"
            else:
                # no api changes
                return f"{prefix}{major}.{minor}.{patch}"


def filter_commits(commits, start=None, end=None):
    """
    Returns commits from start (exclusive) to end (inclusive).
    start must occur before end
    start and end can be a tag, or hash
    """
    if start is not None:
        for idx, c in enumerate(commits):
            if start in c['tags'] or c['hash'].startswith(start):
                commits = commits[:idx]
                break

    if end is not None:
        for idx, c in enumerate(commits):
            if end in c['tags'] or c['hash'].startswith(end):
                commits = commits[idx:]
                break

    return commits


def compile_log(commits):
    """
    """

    versions = []

    # iterate through commits from latest to earliest

    # group by version
    for commit in commits:
        # make a new version if required
        if len(commit['tags']) > 0 or len(versions) == 0:
            tag = commit['tags'][0] if len(commit['tags']) > 0 else 'HEAD'
            versions.append(Version(ref=tag, date=commit['date'],))

        this_commit = Commit(commit)

        # append to current version
        versions[-1].commits.append(this_commit)

        # check if commit is breaking, feature, or fix
        if this_commit.is_breaking:
            versions[-1].breaking += 1

        if this_commit.is_feature:
            versions[-1].feature += 1

        if this_commit.is_fix:
            versions[-1].fix += 1

    # for version in versions:
    #     print(version)

    return versions


def format_log(versions):

    headings = {
        'docs': "Docs",
        'feat': "Features",
        'fix': "Fixes",
        'perf': "Performance",
        'refactor': "Refactorings",
    }

    output = "# Changelog\n\n"

    for version in versions:
        output += f"\n## {version.ref} ({version.date.isoformat()[:10]})\n\n"

        for key, group in groupby(
            sorted(version.commits, key=lambda x: x.type), lambda x: x.type
        ):
            if key in headings:
                output += f"### {headings[key]}\n\n"

                for commit in sorted(group, key=lambda x: f"{x.scope} {x.description}"):
                    scope = f"{commit.scope} - " if commit.scope else ''
                    desc = commit.description
                    desc = sentence(desc) if len(scope) == 0 else desc
                    breaking = " [BREAKING]" if commit.is_breaking else ''
                    output += f"* {sentence(scope)}{desc}{breaking}\n"

                # add following newline
                output += "\n"

    return output


def write_log(log, filename):
    with open(filename, 'w') as f:
        f.write(log)


def main():
    start = 'v0.4.0'
    end = None
    commits = get_commits()
    commits_filtered = filter_commits(commits, start, end)
    versions = compile_log(commits_filtered)

    versions[0].ref = retrieve_current_version(compile_log(commits), prefix='v')
    # print(versions[0].ref)

    log = format_log(versions)
    write_log(log, "CHANGELOG.md")


if __name__ == '__main__':
    main()



from whathappened.git_commits import get_commits


with open('tests/git-log.out') as f:
    test_log = f.read().encode("utf-8")


def mock_check_output(args, **kwargs):
    return test_log


def test_get_commits(monkeypatch):
    monkeypatch.setattr('subprocess.check_output', mock_check_output)

    commits = get_commits()

    print(commits)

    expected = [
        {
            'hash': 'e324c324df48a76113ad9b3c0887f161324046e4',
            'tags': ['v0.1.1'],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 17:30:25 2020 +0200',
            'message': '',
            'title': 'breaking feat(readme): specify expected message format',
        },
        {
            'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
            'tags': [],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 15:00:48 2020 +0200',
            'message': '',
            'title': 'breaking fix (code): repair things',
        },
        {
            'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
            'tags': [],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 15:00:48 2020 +0200',
            'message': '',
            'title': 'breaking fix: repair things',
        },
        {
            'hash': '7b4e7e657f9e3f2f4033cc5f47bcc637f5799fe9',
            'tags': [],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 15:00:48 2020 +0200',
            'message': '',
            'title': 'fix: add inspiration',
        },
        {
            'hash': 'f60445bba0ac48e12ce6be5526644037234ae500',
            'tags': ['v0.0.1'],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 15:00:31 2020 +0200',
            'message': '',
            'title': 'docs (readme): add badges',
        },
        {
            'hash': '9e57ba91f54244af913931c017480a39605c15f9',
            'tags': [],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 13:55:04 2020 +0200',
            'message': (
                'Setup Python 3.6 on ubuntu-latest\n'
                'Install pipenv and dependencies\n'
                'Test'
            ),
            'title': 'build(actions): create python-app.yml for github actions',
        },
        {
            'hash': '4094d22846daea951c4fe0d74abb2a798e9a3404',
            'tags': ['v0.0.0'],
            'author': 'Rollcloud <Rollcloud@users.noreply.github.com>',
            'date': 'Sat Oct 17 13:19:28 2020 +0200',
            'message': '',
            'title': 'Initial commit',
        },
    ]

    assert commits == expected



import subprocess

import setuptools
import versioneer

# check that all tests pass
make_process = subprocess.run(["make", "check"], stderr=subprocess.STDOUT)
if make_process.returncode != 0:

    class TestsException(Exception):
        pass

    raise TestsException("please pass all tests before releasing")

# check that linting passes
make_process = subprocess.run(["make", "lint"], stderr=subprocess.STDOUT)
if make_process.returncode != 0:

    class LintException(Exception):
        pass

    raise LintException("please complete code linting before releasing")

# check to prevent dirty uploads to PyPI
if versioneer.get_versions()["dirty"]:

    class MustCommitError(Exception):
        pass

    raise MustCommitError("please commit everything before releasing")

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ['click']

test_requirements = ['pytest', 'pytest-cov', 'pytest-watch', 'pytest-reportlog']

setuptools.setup(
    author="Rollcloud",
    author_email="roullcloud@gmail.com",
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    cmdclass=versioneer.get_cmdclass(),
    description="A changelog generator using simply structured git commit messages.",
    entry_points={'console_scripts': ['whathappened=whathappened.command_line:cli']},
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="whathappened",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    test_suite='tests',
    tests_require=test_requirements,
    url="https://github.com/Rollcloud/whathappened",
    version=versioneer.get_version(),
)



import click

try:
    from whathappened import __version__, changelog
except ImportError:  # for development use
    import changelog

    __version__ = 'major.minor.patch'


def main(output="CHANGELOG.md", emoji=False, prefix="", git_log_args=[]):
    commits = changelog.get_commits(git_log_args=git_log_args)
    versions = changelog.compile_log(commits)
    versions = changelog.update_latest_version(versions, prefix=prefix)
    log = changelog.format_log(versions, emoji=emoji)

    if output is not None:
        changelog.write_log(log, output)
    else:
        click.echo(log)


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option(
    '--output',
    '-o',
    type=click.Path(),
    default=None,
    help="The changelog destination [default: stdout]",
)
@click.option(
    '--emoji',
    '-e',
    is_flag=True,
    default=False,
    help="Include emoji in headings if present",
)
@click.option(
    '--prefix',
    '-p',
    default="",
    help="Version prefix, often 'version' or 'v' [default: '']",
)
@click.argument('git_log_args', nargs=-1, type=click.UNPROCESSED)
@click.version_option(version=__version__)
def cli(output, emoji, prefix, git_log_args):
    """
    Handle command line arguments. Extra arguments are passed to 'git log'.
    """
    main(output=output, emoji=emoji, prefix=prefix, git_log_args=git_log_args)


if __name__ == '__main__':
    cli()



# https://gist.github.com/simonw/091b765a071d1558464371042db3b959#file-get_commits-py

import subprocess
import re

leading_4_spaces = re.compile('^    ')


def get_commits(git_log_args=()):
    lines = (
        subprocess.check_output(
            ['git', 'log', '--decorate'] + list(git_log_args), stderr=subprocess.STDOUT
        )
        .decode("utf-8")
        .split('\n')
    )
    commits = []
    current_commit = {}

    def save_current_commit():
        title = current_commit['message'][0]
        message = current_commit['message'][1:]
        if message and message[0] == '':
            del message[0]
        current_commit['title'] = title
        current_commit['message'] = '\n'.join(message)
        commits.append(current_commit)

    for line in lines:
        if not line.startswith(' '):
            if line.startswith('commit '):
                if current_commit:
                    save_current_commit()
                    current_commit = {}
                sections = line.split(' ', 2)
                current_commit['hash'] = sections[1]
                current_commit['tags'] = []
                try:
                    references = sections[2][1:-1]  # drop brackets
                    references = references.split(', ')

                    for ref in references:
                        if ref.startswith('tag'):
                            current_commit['tags'].append(ref.split(' ')[1])
                except IndexError:  # if commit has no references
                    pass
            else:
                try:
                    key, value = line.split(':', 1)
                    current_commit[key.lower()] = value.strip()
                except ValueError:
                    pass
        else:
            current_commit.setdefault('message', []).append(
                leading_4_spaces.sub('', line)
            )
    if current_commit:
        save_current_commit()
    return commits



import re

from datetime import datetime
from itertools import groupby

# imported to make this module the base module for all external calls to functions
try:
    from whathappened.git_commits import get_commits  # noqa
except ImportError:  # for development use
    from git_commits import get_commits  # noqa

# https://semver.org/spec/v2.0.0.html
semver_regex = re.compile(
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
    r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
    r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


class VersionFormatException(Exception):
    '''Raise when a version string can not be correctly parsed'''

    pass


class Version:
    def __init__(self, ref, date):
        self.ref = ref
        self.date = (
            datetime.strptime(date, "%a %b %d %H:%M:%S %Y %z")
            if isinstance(date, str)
            else date
        )  # str input format Wed Apr 22 18:58:54 2020 +0200

        self.breaking = 0
        self.feature = 0
        self.fix = 0

        self.commits = []

    def __repr__(self):
        return (
            f"Version('{self.ref}',"
            f" {self.date},"
            f" breaking={self.breaking},"
            f" feature={self.feature},"
            f" fix={self.fix},"
            f" num_commits={len(self.commits)})"
        )  # pragma: no cover


class Commit:

    commit_regex = re.compile(
        r"(?:(?P<breaking>break(?:ing)?)? ?(?P<type>\w+){1}"
        r" ?(?:\((?P<scope>[^\(\):]+)\))?: (?P<description>.+)|"
        r"(?P<description_alt>.+))",
        flags=re.IGNORECASE,
    )

    def __init__(self, commit_dict):
        self.commit_dict = commit_dict

    def __getattr__(self, name):

        try:
            return self.commit_dict[name]
        except KeyError:

            result = Commit.commit_regex.match(self.title)
            breaking, commit_type, scope, description, description_alt = (
                result.group('breaking'),
                result.group('type'),
                result.group('scope'),
                result.group('description'),
                result.group('description_alt'),
            )

            if name == 'description':
                return description if description is not None else description_alt
            elif name == 'is_breaking':
                return breaking is not None
            elif name == 'type':
                return commit_type if commit_type is not None else 'other'
            elif name == 'scope':
                return scope
            elif name == 'is_feature':
                return "feat" in self.type.lower()
            elif name == 'is_fix':
                return "fix" in self.type.lower()
            else:
                raise AttributeError(f"Attribute '{name}' not found in class Commit")

    def __repr__(self):
        return (
            f"Commit({{"
            f"'hash': '{self.hash[:6]}', "
            f"'title': '{self.title}', "
            f"}})"
        )  # pragma: no cover


def sentence(string):
    try:
        return string[0].upper() + string[1:]
    except IndexError:
        # zero-length string
        return string


def calculate_next(versions, prefix=""):
    global semver_regex

    try:
        previous = versions[1]
    except IndexError:
        # if no previous version has been found
        return f"{prefix}0.1.0"

    previous_version = previous.ref[len(prefix) :]
    result = semver_regex.match(previous_version)

    # extract version numbers if possible:
    try:
        major, minor, patch = (
            int(result.group('major')),
            int(result.group('minor')),
            int(result.group('patch')),
        )
    except AttributeError:
        raise VersionFormatException(
            f"The version number of '{previous_version}' with prefix='{prefix}' cannot"
            f" be parsed. Please enter the appropriate prefix or use a version string"
            f" like 'X.Y.Z'. See https://semver.org/spec/v2.0.0.html for more details."
        )

    latest_version = versions[0]

    if major == 0:
        # this is a development release and only receives minor and patch increments
        if latest_version.breaking > 0:
            return f"{prefix}{major}.{minor+1}.0"
        elif latest_version.feature > 0 or latest_version.fix > 0:
            return f"{prefix}{major}.{minor}.{patch+1}"
        else:
            # no api changes
            return f"{prefix}{major}.{minor}.{patch}"

    else:
        # this is production release and receives major, minor, and patch increments
        if latest_version.breaking > 0:
            return f"{prefix}{major+1}.0.0"
        elif latest_version.feature > 0:
            return f"{prefix}{major}.{minor+1}.0"
        elif latest_version.fix > 0:
            return f"{prefix}{major}.{minor}.{patch+1}"
        else:
            # no api changes
            return f"{prefix}{major}.{minor}.{patch}"


def compile_log(commits):
    """
    """

    versions = []

    # iterate through commits from latest to earliest

    # group by version
    for commit in commits:
        # make a new version if required
        if len(commit['tags']) > 0 or len(versions) == 0:
            tag = commit['tags'][0] if len(commit['tags']) > 0 else 'HEAD'
            versions.append(Version(ref=tag, date=commit['date'],))

        this_commit = Commit(commit)

        # append to current version
        versions[-1].commits.append(this_commit)

        # check if commit is breaking, feature, or fix
        if this_commit.is_breaking:
            versions[-1].breaking += 1

        if this_commit.is_feature:
            versions[-1].feature += 1

        if this_commit.is_fix:
            versions[-1].fix += 1

    # for version in versions:
    #     print(version)

    return versions


def update_latest_version(versions, prefix=""):
    '''update the HEAD reference to show the next semver version'''
    latest_version = versions[0]
    latest_version.ref = calculate_next(versions, prefix=prefix)

    return versions


def format_log(versions, emoji=False):
    output = "# Changelog"

    headings = {
        'docs': "Docs",
        'feat': "Features",
        'fix': "Fixes",
        'perf': "Performance",
        'refactor': "Refactorings",
        'other': "Other",
    }

    for version in versions:
        output += f"\n\n## {version.ref} ({version.date.isoformat()[:10]})\n"

        for key, group in groupby(
            sorted(version.commits, key=lambda x: x.type), lambda x: x.type
        ):
            if key in headings:
                if emoji:
                    raise NotImplementedError

                output += f"\n### {headings[key]}\n\n"

                for commit in sorted(group, key=lambda x: f"{x.scope} {x.description}"):
                    scope = f"{commit.scope} - " if commit.scope else ''
                    desc = commit.description
                    desc = sentence(desc) if len(scope) == 0 else desc
                    breaking = " [BREAKING]" if commit.is_breaking else ''
                    output += f"* {sentence(scope)}{desc}{breaking}\n"

    return output


def write_log(log, filename):
    with open(filename, 'w') as f:
        f.write(log)
